#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import re, string, unicodedata

MAP_TO_INTERNATIONAL = {
    # Upper case
    "Á": "A",
    "À": "A",
    "Ä": "A",
    "Ć": "C",
    "Ç": "C",
    "Đ": "D",
    "É": "E",
    "Ê": "E",
    "È": "E",
    "Ë": "E",
    "Í": "I",
    "Î": "I",
    "Ï": "I",
    "Ł": "L",
    "Ö": "O",
    "Ô": "O",
    "Û": "U",
    "Ü": "U",
    "Š": "S",
    # Lower case
    "á": "a",
    "à": "a",
    "ä": "a",
    "ć": "c",
    "ç": "c",
    "đ": "d",
    "é": "e",
    "ê": "e",
    "è": "e",
    "ë": "e",
    "í": "i",
    "î": "i",
    "ï": "i",
    "ł": "l",
    "ö": "o",
    "ô": "o",
    "û": "u",
    "ü": "u",
    "š": "s",
    # Only lower
    "ß": "ss",
}

def to_international_chr(c :chr) -> chr:
    return MAP_TO_INTERNATIONAL.get(c, c)

def to_international_string(s :str) -> str:
    translator = str.maketrans(MAP_TO_INTERNATIONAL)
    return s.translate(translator)

def remove_ponctuation(s :str) -> str:
    translator = str.maketrans({key: None for key in string.punctuation})
    return s.translate(translator)

def to_canonic_string(s :str) -> str:
    s = s.lower()
    s = to_international_string(s)
    s = remove_ponctuation(s)
    s = re.sub("[$^,?!.:]", "", s)
    s = re.sub("\\[a-zA-Z0-9]+ ", "", s) # Remove LateX macros
    s = re.sub(" +", " ", s)
    s = s.replace(" ", "")               # Remove extra spaces
    return s

def to_canonic_fullname(s :str) -> str:
    s = to_international_string(s.lower())
    s = s.replace("-", " ")
    s = re.sub("\(.+\)", "", s)      # Remove surnames
    s = re.sub("\&.*;", "", s)       # Remove HTML
    s = re.sub("[a-zA-Z]+\.", "", s) # Remove sigles
    s = " ".join(s.split())          # Remove useless spaces
    return s

def unicode_to_utf8(s :str) -> str:
    return unicodedata.normalize("NFKD", s)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

class ValuesFromDictFonctor:
    def __init__(self, attributes :list):
        self.m_attributes = attributes

    @property
    def attributes(self) -> list:
        return self.m_attributes

    def __call__(self, entry :dict) -> tuple:
        return tuple(entry.get(k) for k in self.attributes)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from copy                   import deepcopy

from .binary_predicate      import BinaryPredicate
from .connector             import Connector
from .query                 import Query

def rename_list(l :list, mapping :dict) -> list:
    for i, value in enumerate(l):
        try:
            l[i] = mapping[value]
        except KeyError as e:
            #l[i] = None
            pass # see test_dblp.py
    return l

def rename_key(d :dict, old_key :str, new_key :str) -> dict:
    try:
        d[new_key] = d.pop(old_key)
    except KeyError:
        pass
    return d

def rename_entry(d :dict, mapping :dict) -> dict:
    for old_key, new_key in mapping.items():
        rename_key(d, old_key, new_key)
    return d

def rename(mapping :dict, entries :list) -> list:
    for entry in entries:
        rename_entry(entry, mapping)
    return entries

def rename_filters(filters, mapping :dict):
    if filters == None or isinstance(filters, int) or isinstance(filters, tuple):
        pass
    elif isinstance(filters, str):
        try:
            filters = mapping[filters]
        except KeyError:
            pass
    elif isinstance(filters, BinaryPredicate):
        filters.m_left  = rename_filters(filters.left,  mapping)
        filters.m_right = rename_filters(filters.right, mapping)
    else:
        raise RuntimeError("rename_filters: unsupported type %s: filters = %s" % (type(filters), filters))
    return filters

def rename_sort_by(sort_by :dict, mapping :dict) -> dict:
    return {mapping.get(attribute, attribute) : sort_asc for attribute, sort_asc in sort_by.items()} if sort_by else dict()

def rename_query(q :Query, mapping :dict) -> Query:
    q_renamed = q
    rename_list(q_renamed.attributes, mapping)
    rename_filters(q_renamed.filters, mapping)
    q.sort_by = rename_sort_by(q_renamed.sort_by, mapping)
    return q_renamed

def reverse_dict(d :dict) -> dict:
    ret = dict()
    for k,v in d.items():
        ret[v] = k
    return ret

class RenameConnector(Connector):
    def __init__(self, mapping :dict = {}, child = None):
        super().__init__()
        self.m_map_rq = mapping
        self.m_map_qr = reverse_dict(mapping)
        self.m_child  = child

    def attributes(self, object :str) -> set:
        return {self.map_rq.get(attribute, attribute) for attribute in  self.m_child.attributes(object)}

    def query(self, q :Query) -> list:
        super().query(q)
        assert self.child != None
        q_renamed = rename_query(deepcopy(q), self.map_qr)
        return self.answer(
            q,
            rename(
                self.map_rq,
                self.child.query(q_renamed)
            )
        )

    @property
    def child(self):
        return self.m_child

    @property
    def map_qr(self):
        return self.m_map_qr

    @property
    def map_rq(self):
        return self.m_map_rq





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector             import Connector
from .hash                  import to_hashable
from .query                 import Query, ACTION_READ
from .values_from_dict      import ValuesFromDictFonctor

def unique_impl(fonctor :ValuesFromDictFonctor, entries :list) -> list:
    ret = list()
    seen_keys = set()
    for entry in entries:
        key = fonctor(entry)
        if len(key) == 1:
            (key,) = key
        key = to_hashable(key)
        if key not in seen_keys:
            ret.append(entry)
            seen_keys.add(key)
    return ret

def unique(attributes :list, entries :list) -> list:
    fonctor = ValuesFromDictFonctor(attributes)
    return unique_impl(fonctor, entries)

class UniqueConnector(Connector):
    def __init__(self, attributes :list, child):
        super().__init__()
        self.m_fonctor = ValuesFromDictFonctor(attributes)
        self.m_child = child

    @property
    def child(self):
        return self.m_child

    def attributes(self, object :str) -> set:
        return self.child.attributes(object)

    def query(self, q :Query) -> list:
        super().query(q)
        return self.answer(
            q,
            unique_impl(
                self.m_fonctor,
                self.m_child.query(q)
            )
        )

    def __str__(self) -> str:
        return "DUP %s" % ", ".join(self.m_fonctor.attributes)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import io, csv
from enum       import IntEnum

from .connector import Connector
from .query     import Query, ACTION_READ, action_to_str
from .log       import Log

class CsvModeEnum(IntEnum):
    FILENAME = 0 # Use text file
    STRING   = 1 # Use str
    TEXTIO   = 2 # Use any class inheriting TextIOBase

    def __str__(self) -> str:
        s = IntEnum.__str__(self)
        i = len(self.__class__.__name__)
        return s[i + 1:]

class CsvConnector(Connector):
    def __init__(
        self,
        data      :str,        # filename | csv_str | TextIOBase
        delimiter :chr         = ' ',
        quotechar :chr         = '"',
        mode      :CsvModeEnum = CsvModeEnum.FILENAME
    ):
        super().__init__()
        stream = open(data, "rt")  if mode == CsvModeEnum.FILENAME else \
                 io.StringIO(data) if mode == CsvModeEnum.STRING   else \
                 data              if mode == CsvModeEnum.TEXTIO   else \
                 None
        if not stream:
            raise RuntimeError("CsvConnector: Invalid input stream (data = %r, mode = %s)" % (data, mode))

        # csv.reader cannot directly process a str, so we use a io.StringIO
        # See https://stackoverflow.com/questions/3305926/python-csv-string-to-array
        Log.debug("delimiter = %r, quotechar = %r" % (delimiter, quotechar))
        Log.debug(type(stream))
        reader = csv.reader(stream, delimiter=delimiter, quotechar=quotechar)
        rows = [row for row in reader]
        Log.debug(rows)

        if mode == CsvModeEnum.FILENAME and stream:
            stream.close()

        # Assuming that attributes are declared in the first line of the CSV data.
        self.indexed_attributes = rows[0]
        self.entries = [{self.indexed_attributes[i] : v for i, v in enumerate(row)} for row in rows[1:]]

    def attributes(self, object :str):
        return set(self.indexed_attributes)

    def query(self, query :Query) -> list:
        super().query(query)
        ret = list()
        if query.action == ACTION_READ:
            ret = self.reshape_entries(query, self.entries)
        else:
            raise RuntimeError("CsvConnector.query: %s not yet implemented" % action_to_str(query.action))
        return self.answer(query, ret)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector             import Connector
from .query                 import Query, ACTION_READ
from .values_from_dict      import ValuesFromDictFonctor

def sort_by_impl(fonctor : ValuesFromDictFonctor, entries :list, desc = True) -> list:
    return sorted(entries, key = fonctor, reverse = desc)

def sort_by(attributes :list, entries :list, desc = False) -> list:
    fonctor = ValuesFromDictFonctor(attributes)
    return sort_by_impl(fonctor, entries, desc)

class SortByConnector(Connector):
    def __init__(self, attributes :list, child, desc = False):
        super().__init__()
        self.m_fonctor = ValuesFromDictFonctor(attributes)
        self.m_child = child
        self.m_desc = desc

    def attributes(self, object :str) -> set:
        return self.child.attributes(object)

    @property
    def child(self):
        return self.m_child

    @property
    def desc(self) -> bool:
        return self.m_desc

    def query(self, q :Query) -> list:
        super().query(q)
        return self.answer(
            q,
            sort_by_impl(
                self.m_fonctor,
                self.m_child.query(q),
                self.m_desc
            )
        )

    def __str__(self) -> str:
        return "SORT BY %s %s" % (
            ", ".join(self.m_fonctor.attributes),
            "DESC" if self.desc else "ASC"
        )




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# Usage:
#     python3 setup.py install
#     python3 setup.py bdist_rpm

from setuptools import find_packages, setup

__version__ = (0, 9, 3)

README, HISTORY = "", ""
try:
    with open("README.rst") as f_readme:
        README = f_readme.read()
except:
    pass

try:
    with open("HISTORY.rst") as f_history:
        HISTORY = f_history.read()
except:
    pass

setup(
    name             = "minifold",
    version          = ".".join(["%s" % x for x in __version__]),
    description      = "minifold",
    long_description = "%s\n\n%s" % (README, HISTORY),
    long_description_content_type = "text/x-rst",
    author           = "Marc-Olivier Buob",
    author_email     = "marc-olivier.buob@nokia-bell-labs.com",
    url              = "http://github.com/nokia/minifold",
    license          = "BSD-3",
    packages         = find_packages(),
    package_dir      = {"minifold" : "minifold"},
    python_requires  = ">=3",
    requires         = ["typing"],
    test_suite       = "tests",
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob, Dylan Sellar"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"


class ForEachFilter():
    def __init__(self, attribute_name: str, map_filter: dict):
        self.attribute_name = attribute_name
        self.map_filter = map_filter

    def match_filters(self, sub_entry):
        for attribute, filt in self.map_filter.items():
            if attribute in sub_entry.keys():
                if not filt(sub_entry[attribute]):
                    return False
        return True

    def __call__(self, entry: dict) -> bool:
        sub_entries = entry.get(self.attribute_name)

        if not sub_entries:
            return True

        entry[self.attribute_name] = [
            sub_entry \
            for sub_entry in sub_entries \
            if self.match_filters(sub_entry)
        ]

        return True

def for_each_sub_entry(entry, attribute :str, map_lambda :dict) -> dict:
    entry[attribute] = [
        {
              k : map_lambda[k](v) if k in map_lambda.keys() else v for k, v in sub_entry.items()
        } for sub_entry in entry[attribute]
    ]
    return entry




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from minifold.log import Log

try:
    import pycountry
except ImportError as e:
    Log.warning("Please install pycountry: apt-get install python3-pycountry")
    raise e

def _country_code_to_name(country_code :str) -> str:
    # For obsolete version of pycountry
    Log.warning("Please update python3-pycountry; apt-get update && apt-get upgrade")

    ret = None

    try:
        country = pycountry.countries.get(alpha_2 = country_code.upper())
        ret = country.name
    except KeyError as e:
        Log.warning("Unknown country %r" % country_code)

    return ret

def country_code_to_name(country_code :str) -> str:
    if country_code == None:
        return None

    ret = None
    try:
        country = pycountry.countries.get(alpha_2 = country_code.upper())
        if country is None:
            return None
        ret = country.name
    except KeyError as e:
        if "%s" % e == "alpha_2":
            ret = _country_code_to_name(country_code)
        else:
            Log.warning("Unknown country %r" % country_code)

    return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import sys
from pprint     import pformat
from .query     import Query
from .log       import Log

class Connector:
    trace_queries = False
    trace_entries = False
    trace_only_keys = False

    def __init__(self):
        """
        Constructor.
        """
        pass

    if sys.version_info >= (3, 6):
        subclasses = dict()

        #https://stackoverflow.com/questions/5189232/how-to-auto-register-a-class-when-its-defined
        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__(**kwargs)
            name = repr(cls).split("'")[1]
            Connector.subclasses[name] = cls

        @staticmethod
        def get_class(name :str):
            cls = Connector.subclasses.get(name)
            if not cls:
                raise RuntimeError(
                    "Invalid connector type [%s]. Known types are:\n%s\n" % (
                        name,
                        "\n\t".join([str(name) for name in subclasses.keys()])
                    )
                )
            return cls

    def query(self, query :Query) -> list:
        """
        Method called when this Connector handle an input Query.
        Args:
            query: The handled query.
        Returns:
            The list of entries matching the input Query.
        """
        if Connector.trace_queries:
            Log.debug("%r: --> %s" % (self, query))
        # This method must be overloaded in child classes as follows:
        # super().query(query)
        # entries = [ ... ]
        # return self.answer(query, entries)
        return list()

    def attributes(self, object :str) -> set:
        """
        List available attributes related to a given object.
        Args:
            object: The name of the object
        Returns:
            The set of available attributes for object.
        """
        # Must be overwritten in child class if reshape_entries is needed
        raise NotImplemented

    def reshape_entries(self, query :Query, entries :list) -> list:
        """
        Reshape entries returned by self.query() before calling self.answer().
        This method should only be called if the Connector only support a subset
        of Query operators among {SELECT, WHERE, LIMIT, OFFSET}
        Args:
            query: The handled Query instance.
            entries: The list of raw entries fetched so far, corresponding to
                SELECT * FROM foo LIMIT n
                where n >= query.limit
        """
        max_attributes = self.attributes(query.object)
        attributes = set(query.attributes) & max_attributes if query.attributes else max_attributes

        ret = list()
        for entry in entries[query.offset:]:
            # LIMIT
            if len(ret) == query.limit:
                break

            # WHERE
            if query.filters is None or query.filters(entry):
                # SELECT
                entry = {k : v for (k, v) in entry.items() if k in attributes}
                missing_attributes = set(attributes) - set(entry.keys())
                for k in missing_attributes:
                    entry[k] = None

                ret.append(entry)
        return ret

    def answer(self, query :Query, ret):
        """
        Method traversed when this Connector is ready to answer to a given Query.
        Args:
            query: The related Query instance.
            ret: The corresponding result.
        """
        if Connector.trace_entries or Connector.trace_only_keys:
            if isinstance(ret, list) and Connector.trace_only_keys and len(ret) > 0:
                entry = ret[0]
                message = "Forwarding %d entries, keys = {%s}" % (
                    len(ret),
                    ", ".join(entry.keys())
                )
            else:
                message = pformat(ret)
            Log.debug("%r: <-- %s\n%s" % (self, query, message))
        return ret



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

import json, sys
from minifold.connector import Connector
from minifold.singleton import Singleton

# Example of configuration

DEFAULT_MINIFOLD_CONFIG = """{
    "dblp:dagstuhl" : {
        "type" : "minifold.dblp.DblpConnector",
        "args" : {
            "dblp_api_url" : "https://dblp.dagstuhl.de"
        }
    },
    "dblp:uni-trier" : {
        "type" : "minifold.dblp.DblpConnector",
        "args" : {
            "dblp_api_url" : "https://dblp.uni-trier.de"
        }
    }
}"""

# Usage:
# from minifold.dblp import DblpConnector
# config = Config()
# config.loads(DEFAULT_MINIFOLD_CONFIG)
# dblp1 = config.make_connector("dblp:dagstuhl")
# dblp2 = config.make_connector("dblp:uni-trier")

class Config(dict, metaclass=Singleton):
    def loads(self, s_json :str):
        self.update(json.loads(s_json))

    def load(self, stream):
        self.update(json.load(stream))

    def load_file(self, filename :str):
        with open(filename, "r") as f:
            self.load(f)

    def make_connector(self, name :str):
        conf = self.get(name)
        if not conf:
            raise KeyError(
                "Config: Key [%s] not found. Known configuration are:\n%s\n" % (
                    name,
                    "\n\t".join([str(k) for k in self.keys()])
                )
            )
        type = conf["type"]
        cls = Connector.subclasses.get(type)
        if not cls:
            raise KeyError(
                "Config: Connector [%s] not found. Known connectors are:\n%s\n" % (
                    type,
                    "\n\t".join([str(k) for k in Connector.subclasses.keys()])
                )
            )
        kwargs = conf.get("args", dict())
        return cls(**kwargs)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

# Based on https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

class Singleton(type):
    s_instances = dict()

    def __call__(cls, *args, **kwargs):
        if cls not in cls.s_instances:
            cls.s_instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.s_instances[cls]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# If you need more generic join, see join_if.py
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector import Connector
from .query     import Query, ACTION_READ
from .join_if   import merge_dict

def are_naturally_joined(l :dict, r :dict) -> bool:
    inter_keys = set(l.keys()) & set(r.keys())
    if len(inter_keys) == 0:
        return False
    for key in inter_keys:
        if l[key] != r[key]:
            return False
    return True

def natural_join(l_entries :list, r_entries :list) -> list:
    ret = list()
    for l in l_entries:
        for r in r_entries:
            if are_naturally_joined(l, r):
                ret.append(merge_dict(l, r))
    return ret

class NaturalJoinConnector(Connector):
    def __init__(self, left, right):
        self.m_left = left
        self.m_right = right
        self.m_left_entries = list()
        self.m_right_entries = list()

    def attributes(self, object :str) -> set:
        return self.m_left.attributes(object) \
             | self.m_right.attributes(object)

    def query(self, query :Query) -> list:
        self.m_left_entries.clear()
        self.m_right_entries.clear()
        self.m_left_entries = self.left.query(query)
        self.m_right_entries = self.right.query(query)
        return self.answer(
            query,
            natural_join(self.m_left_entries, self.m_right_entries)
        )

    @property
    def left(self):
        return self.m_left

    @property
    def right(self):
        return self.m_right





#!/usr/bin/env python3
#
# Twitter connector
#
# Authors:
#   Fabien Mathieu    <fabien.mathieu@nokia-bell-labs.com>
#   Marc-Olivier Buob <marc-olivier.buob@nokia-bell-labs.com>

import datetime, tweepy
from minifold.connector import Connector
from minifold.log       import Log
from minifold.query     import ACTION_READ, Query

def tweet_to_dict(tweet):
    # Fetch tweet
    content = None
    try:
        content = tweet._json["full_text"]
    except:
        content = tweet._json["text"]
    if content:
        content = content.split(" https://t.co/")[0]

    # Get tweet image
    image = None
    try:
        image = tweet._json["entities"]["media"][0]["media_url"]
    except:
        # Try to get image from youtube, e.g: https://img.youtube.com/vi/iZCyv0e2RLM/1.jpg
        try:
            if "youtu" in tweet._json["entities"]["urls"][0]["display_url"]:
                image = "https://img.youtube.com/vi/%s/1.jpg" % tweet._json["entities"]["urls"][0]["display_url"].split("/")[-1]
        except:
            Log.warning("No image found for tweet %s" % content)

    # Get author name
    author_name = None
    try:
        author_name = tweet._json["user"]["name"]
    except:
        Log.warning("No author name found for tweet %s" % content)

    # Get author image
    author_image = None
    try:
        author_image = tweet._json["user"]["profile_image_url"]
    except:
        Log.warning("No author image found for tweet %s" % content)

    # Get tweet date
    date_list = tweet._json["created_at"].split()
    display_date = "%(m)s %(d)s, %(y)s" % {
        "d" : date_list[2],
        "m" : date_list[1],
        "y" : date_list[5]
    }
    date = datetime.datetime.strptime(display_date, "%b %d, %Y")
    display_date = date.strftime("%d/%m/%Y")

    # Return entry
    return {
        "text"         : content,
        "image"        : image,
        "date"         : date,
        "display_date" : display_date,
        "author_name"  : author_name,
        "author_image" : author_image,
    }

class TwitterConnector(Connector):
    def __init__(self, twitter_id :str, consumer_key :str, consumer_secret :str, access_token :str, access_token_secret :str):
        """
        Constructor.
        """
        self.twitter_id          = twitter_id
        self.consumer_key        = consumer_key
        self.consumer_secret     = consumer_secret
        self.access_token        = access_token
        self.access_token_secret = access_token_secret
        self.api                 = None
        self.connect()

    def attributes(self, object :str) -> set:
        ret = {
            "text",
            "image",
            "date",
            "display_date",
            "author_name",
            "author_image"
        } if object in {"self", "feed"} else set()

    def connect(self):
        """
        Connect to twitter API.
        """
        auth = tweepy.OAuthHandler(self.consumer_key, self.consumer_secret)
        auth.set_access_token(self.access_token, self.access_token_secret)
        self.api = tweepy.API(auth)

    def query(self, q :Query) -> list:
        """
        Query the TwitterConnector.
        Args:
            q: A minifold Query.
        Returns:
            A list of dict containing the queried entries.
        """
        if not self.api:
            raise RuntimeError("Not connected to twitter")

        if q.action == ACTION_READ:
            if q.object in {"self", "feed"}:
                # Retrieve Tweets
                assert q.limit > 0
                size = q.limit
                if q.object == "self":
                    tweets = self.api.user_timeline(id = self.twitter_id, tweet_mode = "extended", count = size)
                elif q.object == "feed":
                    tweets = self.api.home_timeline(id = self.twitter_id, tweet_mode = "extended", count = size)
                entries = [tweet_to_dict(tweet) for tweet in tweets]
            else:
                raise ValueError("TwitterConnector: invalid object '%s'" % q.object)
        else:
            raise RuntimeError("TwitterConnector.query(%s): action %s not implemented" % (q, q.action))
        return self.answer(q, entries)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

def cast_none(s :str):
    """
    Cast a string into None if possible.
    Args:
        s: The input str.
    Raises:
        ValueError: if the cast cannot be achieved.
    Returns:
        None if successful.
    """
    if s is None:
        return None
    elif isinstance(s, str) and s.lower() == "none":
        return None
    raise ValueError("Invalid literal for cast_none(): %s" % s)

def cast_bool(s :str) -> bool:
    """
    Cast a string into a bool if possible.
    Args:
        s: The input str.
    Raises:
        ValueError: if the cast cannot be achieved.
    Returns:
        The boolean corresponding to s if successful.
    """
    if isinstance(s, bool):
        return s
    elif isinstance(s, str):
        l = s.lower()
        if l == "true":
            return True
        elif l == "false":
            return False
    raise ValueError("Invalid literal for cast_bool(): %s" % s)

def lexical_cast(s :str, cast):
    """
    Cast a string according to an operator.
    Args:
        cast: The cast operator.
        s: The string to be casted.
    Raises:
        ValueError: if the cast cannot be achieved.
    Returns:
        The corresponding value.
    """
    return cast(s)

def lexical_casts(s :str, cast_operators = [cast_none, cast_bool, int, float]):
    """
    Cast a string according to several cast operators.
    Args:
        s: The string to be casted.
        cast_operators: A list of cast operators. Operators must be ordered
            from the less strict to the more strict (e.g. int before float).
    Returns:
        The original string if no cast worked, the corresponding casted value
        otherwise.
    """
    for cast in cast_operators:
        try:
            ret = lexical_cast(s, cast)
            return ret
        except ValueError:
            pass
    return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of the minifold project.
# https://github.com/nokia/minifold

__author__     = "Marc-Olivier Buob"
__maintainer__ = "Marc-Olivier Buob"
__email__      = "marc-olivier.buob@nokia-bell-labs.com"
__copyright__  = "Copyright (C) 2018, Nokia"
__license__    = "BSD-3"

from .connector             import Connector
from .query                 import Query, ACTION_READ

def unnest(map_key_unnestedkey :dict, entries :list) -> list:
    ret = list()
    for entry in entries:
        # NOTE: This only works if entries a single key matching map_key_unnestedkey
        for k, values in entry.items():
            new_attribute = map_key_unnestedkey.get(k)
            if isinstance(values, list):
                for value in values:
                    ret.append({new_attribute : value})
            else:
                ret.append({new_attribute : values})
    return ret

class UnnestConnector(Connector):
    def __init__(self, map_key_unnestedkey :dict, child):
        super().__init__()
        self.m_child = child
        self.m_map_key_unnestedkey = map_key_unnestedkey

    def attributes(self, object :str) -> set:
        raise NotImplemented

    @property
    def child(self):
        return self.m_child

    def unnest_key(self, key) -> str:
        return self.m_map_key_unnestedkey[key]

    def query(self, q :Query) -> list:
        super().query(q)
        return self.answer(
            q,
            unnest(
                self.m_map_key_unnestedkey,
                self.child.query(q)
            )
        )


